package nu.zoom.dsl.ast;

import nu.zoom.dsl.parser.EndpointsBaseVisitor;
import nu.zoom.dsl.parser.EndpointsParser;
import org.antlr.v4.runtime.tree.TerminalNode;

import java.util.*;
import java.util.stream.Collectors;

public class EndpointsVisitorTransformer extends EndpointsBaseVisitor<EndpointsParser.DocumentContext> {
    private final ArrayList<EndpointNode> endpoints = new ArrayList<>();
    private final HashMap<String,String> config = new HashMap<>();
    private final HashSet<TypeNode> dataTypes = new HashSet<>();

    public EndpointsVisitorTransformer() {
    }

    public List<EndpointNode> getEndpoints() {
        return List.copyOf(endpoints);
    }

    public Map<String,String> getConfig() {
        return Map.copyOf(config);
    }

    public List<TypeNode> getDataTypes() {
        return List.copyOf(dataTypes);
    }

    @Override
    public EndpointsParser.DocumentContext visitConfigitem(EndpointsParser.ConfigitemContext ctx) {
        String configKey = ctx.configkey().IDENTIFIER().getText();
        String configValue = getText(ctx.configvalue().IDENTIFIER(), ctx.configvalue().VALUE());
        this.config.put(configKey, configValue);
        return super.visitConfigitem(ctx);
    }

    @Override
    public EndpointsParser.DocumentContext visitNamedTypeDeclaration(EndpointsParser.NamedTypeDeclarationContext ctx) {
        var type = extractCompoundTypeNode(ctx);
        addDatatType(type, ctx.start.getLine());
        return super.visitChildren(ctx);
    }

    @Override
    public EndpointsParser.DocumentContext visitEndpoint(EndpointsParser.EndpointContext ctx) {
        List<String> segments = ctx
                .path()
                .pathSegment()
                .stream()
                .map(
                        segment -> getText(segment.IDENTIFIER(), segment.VALUE())
                ).toList();
        extractEndpoint(ctx, segments);
        return super.visitEndpoint(ctx);
    }

    private void extractEndpoint(EndpointsParser.EndpointContext ctx, List<String> segments) {
        EndpointsParser.RequestBodyContext requestBodyContext = ctx.requestBody() ;
        String requestTypeName = null ;
        // request can be either a named type declaration, an anonymous type declaration or a type reference
        if (requestBodyContext.namedTypeDeclaration() != null) {
            var compoundTypeNode = extractCompoundTypeNode(ctx.requestBody().namedTypeDeclaration());
            requestTypeName = compoundTypeNode.name() ;
            // We don't save the datattype, it will be handled by the child visitor that finds a named datatype
        } else if (requestBodyContext.typeDeclaration() != null) {
            var typeFields = extractTypeFields(requestBodyContext.typeDeclaration().typeField()) ;
            requestTypeName = segments.getLast() + "Request" ;
            addDatatType(new TypeNode(requestTypeName, typeFields), ctx.start.getLine());
        } else if (requestBodyContext.IDENTIFIER() != null) {
            requestTypeName = requestBodyContext.IDENTIFIER().getText() ;
        }
        if (requestTypeName == null) {
            throw new ParseException("Unable to create the request body data type for " + segments.stream().collect(Collectors.joining())) ;
        }
        var endpoint = new EndpointNode(new PathsNode(segments), requestTypeName);
        this.endpoints.add(endpoint);
    }

    private TypeNode extractCompoundTypeNode(EndpointsParser.NamedTypeDeclarationContext ctx) {
        String typeName = ctx.typeName().getText();
        List<FieldNode> fields = extractTypeFields(ctx.typeDeclaration().typeField());
        return new TypeNode(typeName, fields);
    }

    private List<FieldNode> extractTypeFields(List<EndpointsParser.TypeFieldContext> compoundFieldContexts) {
        return compoundFieldContexts.stream().map(
                ctx -> new FieldNode(
                        ctx.fieldName().getText(),
                        ctx.fieldType().getText()
                )
        ).toList();
    }

    // Concatenate the text from to terminal nodes. Useful for contexts that are either an identifier or a value,
    // and you just want the text from whichever is not null.
    private String getText(TerminalNode identifier, TerminalNode value) {
        return
                ((identifier != null) ? identifier.getText() : "") +
                        ((value != null) ? value.getText() : "");
    }

    private void addDatatType(TypeNode type, int lineNumber) throws ParseException {
        if (this.dataTypes.stream().filter(t -> t.name().equals(type.name())).findAny().isPresent()) {
            throw new ParseException("Duplicate datatype '" + type.name()+ "' at line " + lineNumber) ;
        } else {
            this.dataTypes.add(type);
        }
    }
}
