package nu.zoom.dsl.cli;

import nu.zoom.dsl.ast.DocumentNode;
import nu.zoom.dsl.ast.ParserWrapper;
import nu.zoom.dsl.jte.Generator;
import picocli.CommandLine;
import picocli.CommandLine.Command;
import picocli.CommandLine.Option;
import picocli.CommandLine.Parameters;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.concurrent.Callable;

@Command(
	name = "EndpointsCLI",
	mixinStandardHelpOptions = true,
	description = "Generate source code from an endpoints specification file."
)
public class EndpointsCLI implements Callable<Integer> {
	@SuppressWarnings("unused")
    @Parameters(index = "0", description = "The source endpoints DSL file.")
	private Path file;

	@SuppressWarnings("CanBeFinal")
    @Option(names = {"-t", "--template"}, description = "The template directory. Default is ~/endpoints-templates")
	private Path templateDir = Paths.get(System.getProperty("user.dir"), "endpoints-templates");

	@SuppressWarnings("CanBeFinal")
    @Option(names = {"-o", "--output"}, description = "The directory to write the generated code to. Default is ~/endpoints-output")
	private Path outputDir = Paths.get(System.getProperty("user.dir"), "endpoints-output");

	@SuppressWarnings("unused")
    @Option(names = {"-v", "--verbose"}, description = "Print verbose debug messages.")
	private Boolean verbose = false;

	public static void main(String[] args) {
		int exitCode = new CommandLine(new EndpointsCLI()).execute(args);
		System.exit(exitCode);
	}

	@Override
	public Integer call() {
		try {
			validateTemplateDirectory();
			validateInputFile();
			validateOutputDirectory();
			DocumentNode rootNode = ParserWrapper.parse(file);
			Generator generator = new Generator(templateDir, rootNode, outputDir);
			List<Path> generatedPaths = generator.generate();
			if (generatedPaths.isEmpty()) {
				System.out.println("No generated paths found.");
			}
			return 0;
		} catch (Exception e) {
			System.err.println(e.getMessage());
			return 1;
		}
	}

	private void validateOutputDirectory() throws IOException {
		if (Files.notExists(this.outputDir)) {
			Files.createDirectories(this.outputDir);
		}
		if (!Files.isDirectory(this.outputDir)) {
			throw new IllegalArgumentException("Output directory: '" + this.outputDir + " 'is not a directory.");
		}
	}

	private void validateTemplateDirectory() {
		if (!Files.isDirectory(this.templateDir)) {
			throw new IllegalArgumentException("Template directory '" + this.templateDir + "' is not a directory.");
		}
	}

	private void validateInputFile() {
		if (Files.notExists(this.file)) {
			throw new IllegalArgumentException("Input file '" + this.file + "' does not exist.");
		}
		if (!Files.isReadable(this.file) || !Files.isRegularFile(this.file)) {
			throw new IllegalArgumentException("Input file '" + this.file + "' is not a readable file.");
		}
	}
}
